/**
 * CRFalign: A CRF-based sentence aligner for parallel corpora.
 * 
 * Copyright (C) 2012  Eva Mujdricza-Maydt
 * Copyright (C) 2012  Huiqin Körkel-Qu
 * 
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, see <http://www.gnu.org/licenses/>.
 * 
 * Developers:
 * - Eva Mujdricza-Maydt (mujdricz@cl.uni-heidelberg.de)
 * - Huiqin Körkel-Qu (huiqin@cl.uni-heidelberg.de)
 */

package annotators.champollion;

import java.io.BufferedReader;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.InputStreamReader;
import java.io.OutputStreamWriter;
import java.util.ArrayList;

/** For Champollion we need a source-target dictionary.
 * We take the Omega german-english FreeDict dictionary. This will be extracted
 * into the format which can be read by Champollion.
 *
 * @author Eva Mujdricza-Maydt
 * @lastChange 20120626
 */
class ExtractDict {
	
	public static void main(String[] args) 
		throws FileNotFoundException, IOException {
		try {
			String origDict = args[0];
			String champollDict = args[1];

			writeOmegaDictInChampollionFormat(origDict, champollDict);
		} catch (ArrayIndexOutOfBoundsException aioe) {
			System.out.println("Usage:\njava annotators/champollion/ExtractDict "
				+ "<Omega dictionary file name with path> "
				+ "<output dictionary file name with path>");
		}
	}
	
	/** Transforms the Omega dictionary in a dictionary format Champollion
	 * can take.
	 * 
	 * @param dictInputFileName the file name (with path) of the Omega dictionary
	 * @param dictOutputFileName file name of the output dictionary (with path)
	 * @throws FileNotFoundException
	 * @throws IOException 
	 */
	private static void writeOmegaDictInChampollionFormat(
		String dictInputFileName,
		String dictOutputFileName) 
		throws FileNotFoundException, IOException {
		
		System.out.println("EXTRACTING: " + dictInputFileName);
		ArrayList<String[]> dict = new ArrayList<String[]>();
		InputStreamReader in = new InputStreamReader(
			new FileInputStream(dictInputFileName), "iso8859-1");

		BufferedReader br = new BufferedReader(in);
		String line = null;
		
		while((line = br.readLine()) != null) {
			if (line.startsWith("#") || line.isEmpty()) {
				 continue;
			}
			String[] array = line.trim().split("\t");
			if (array.length != 2) {
				String msg = "line could not be splitted by tab into 2 parts: " + line;
				throw new ArrayIndexOutOfBoundsException();
			}
			String[] src = array[0].split(" ; ");
			String[] trg = array[1].split(" ; ");
			
			for (String s : src) {
				for (String t : trg) {
					String[] d = new String[2]; 
					d[0] = s;
					d[1] = t;
					dict.add(d);
					System.out.println(s + " -> " + t);
				}
			}
		}
		
		br.close();
		System.out.println("DICTIONARY size: " + dict.size());
		
		System.out.println("WRITING: " + dictOutputFileName);

		OutputStreamWriter fw = new OutputStreamWriter(
			new FileOutputStream(dictOutputFileName), "iso8859-1"); 

		for (String[] d : dict) {
			fw.write(d[0] + " <> " + d[1] + "\n");
			fw.flush();
		}
		
		fw.close();
	}
}
