/**
 * CRFalign: A CRF-based sentence aligner for parallel corpora.
 * 
 * Copyright (C) 2012  Eva Mujdricza-Maydt
 * Copyright (C) 2012  Huiqin Körkel-Qu
 * 
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, see <http://www.gnu.org/licenses/>.
 * 
 * Developers:
 * - Eva Mujdricza-Maydt (mujdricz@cl.uni-heidelberg.de)
 * - Huiqin Körkel-Qu (huiqin@cl.uni-heidelberg.de)
 */
 
package corpora.gutenberg;

import com.cybozu.labs.langdetect.LangDetectException;
import java.io.*;
import java.util.ArrayList;
import objects.Alignment_DocLevel;
import objects.Alignment_SentLevel;
import objects.Sentence;

/** Extracts and organizes the texts of the Gutenberg corpus.
 *
 * @author Eva Mujdricza-Maydt
 * @lastChange 20120626
 */
public class ExtractGutenberg {
	
	/** Rewrites the original Gargantua alignment outputs in OPUS-style format.
	 * (See README.txt).
	 * 
	 * @param alignmentPath path to the Gargantua output alignment files
	 * @param inputExtension extension of the Gargantua output alignment files
	 * @param outputExtension extension for the rewritten files
	 * @throws FileNotFoundException
	 * @throws IOException 
	 */
	public static void writeGargantuaAlignmentsInOpusFormat(
		String alignmentPath,
		String inputExtension, String outputExtension) 
		throws FileNotFoundException, IOException {
		
		ArrayList<String> alignmentFileNames = utils.Helper.extractFileNamesByEnding_woEnding(
			alignmentPath, inputExtension);
		
		for (String fileStem : utils.Helper.sortAL_str(alignmentFileNames)) {
			System.out.println("- " + fileStem);
			
			String currentAlignmentFileName = alignmentPath + fileStem + inputExtension;
			//the source and target files have the same file name
			Alignment_DocLevel docAlign = new Alignment_DocLevel(fileStem, fileStem); 

			ArrayList<int[]> listWith1To1Alignments = corpora.ExtractCorpus_forEvaluation.
				readInPartialAlignments_gargantuaAnnotation(currentAlignmentFileName);

			//now build the sentence-alignments
			ArrayList<Alignment_SentLevel> sentAligns = 
				eval.EvaluateCorpus.buildSentAlignmentsFromPartialAlignments(
				listWith1To1Alignments);

			docAlign.setSentAlignments(sentAligns);
				
			System.out.println("\t" + docAlign.getSentAlignments().size() + " alignments");
			
			//output
			String opusFormatOutput = alignmentPath + fileStem + outputExtension;
			
			corpora.ExtractCorpus_forEvaluation.writeAlignmentFile_opusFormat(
				opusFormatOutput, docAlign);
		}
		System.out.println();
	}

	
	/** Writes the plain text format with tokenized words for Gutenberg files.
	 * <b/>NOTE that for Gutenberg data the original texts are already 
	 * tokenized plain texts.
	 * 
	 * @param pathToDataSet path to the current data set
	 * @param fileStems list of the file names without their file extension
	 * @param extension the extension of the file names
	 * @throws FileNotFoundException
	 * @throws IOException
	 * @throws LangDetectException 
	 */
	public static void writeGutenbergPlainTexts_tokens(
		String pathToDataSet, ArrayList<String> fileStems, String extension) 
		throws FileNotFoundException, IOException, LangDetectException {
		
		for (String fileStem : utils.Helper.sortAL_str(fileStems)) {
			String fileName = fileStem + extension;
			
			ArrayList<Sentence> textInSentences_src = corpora.ExtractCorpus.readInPlainText(
				pathToDataSet + main.Constants_Main.FOLDER_SRC, fileName, true);
			
			String plainTextFileName_src = pathToDataSet + main.Constants_Main.FOLDER_SRC + fileStem + main.Constants_Main.EXT_TEXT_TOKENS;
			
			corpora.ExtractCorpus.writePlainText_tokens(textInSentences_src, plainTextFileName_src);
			ArrayList<Sentence> textInSentences_trg = corpora.ExtractCorpus.readInPlainText(
				pathToDataSet + main.Constants_Main.FOLDER_TRG, fileName, true);
			System.out.print(".");
			String plainTextFileName_trg = pathToDataSet + main.Constants_Main.FOLDER_TRG + fileStem + main.Constants_Main.EXT_TEXT_TOKENS;

			corpora.ExtractCorpus.writePlainText_tokens(textInSentences_trg, plainTextFileName_trg);
			System.out.print(".");
		}
		System.out.println();
	}
	
}
