/**
 * CRFalign: A CRF-based sentence aligner for parallel corpora.
 *
 * Copyright (C) 2012 Eva Mujdricza-Maydt Copyright (C) 2012 Huiqin Körkel-Qu
 *
 * This program is free software: you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free Software
 * Foundation, either version 3 of the License, or (at your option) any later
 * version.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
 * details.
 *
 * You should have received a copy of the GNU General Public License along with
 * this program. If not, see <http://www.gnu.org/licenses/>.
 *
 * Developers: - Eva Mujdricza-Maydt (mujdricz@cl.uni-heidelberg.de) - Huiqin
 * Körkel-Qu (huiqin@cl.uni-heidelberg.de)
 */
package diceMatrix;

import java.io.BufferedReader;
import java.io.BufferedWriter;
import java.io.File;
import java.io.FileReader;
import java.io.FileWriter;
import java.io.IOException;
import java.util.ArrayList;
import java.util.HashMap;

/**
 * change every word in a file from string to be integer.
 *
 * the function changFileTo Array read the text line by line. and for every
 * line, remove the stopwords at first then change the words left to be the
 * corresponding nums and save this sent into a integer array. e.g the seven
 * sentence in the text is
 *
 * I have a dog.
 *
 * "a" is a stop word. "have" has index 7, "I" has index 18 "dog" has index 2879
 *
 * we will get the integer array [18,7,2879] and the 7th element of the output
 * array is [18,7,2879]
 *
 * @author Huiqin Körkel-Qu
 */
public class wordToNum {

	HashMap<String, int[]> multiWordsHash;
	String[] singleWords;
	HashMap<String, Integer> stopWordsList;

	/**
	 *
	 * @param multiWordsHash: the tokens that appear more than ones , the
	 * corresponding int for it, and the times it appear in training data
	 * @param singleWords: the list of tokens that appear only once in training
	 * data.
	 * @param stopWordsList: the stop word list, the stop word will be
	 * eliminated and not be changed to be int.
	 */
	public wordToNum(HashMap<String, int[]> multiWordsHash, String[] singleWords, HashMap<String, Integer> stopWordsList) {
		this.multiWordsHash = multiWordsHash;
		this.singleWords = singleWords;
		this.stopWordsList = stopWordsList;
	}

	/**
	 *
	 * @param file: the raw file in which each line is a sentence
	 * @param numFile: change the word in rawfile to be int, and get a new file
	 * which contains only integers.
	 * @return ArrayList of int[], each int[] represents a sentence, the the
	 * ArrayList is the whole file.
	 * @throws IOException
	 */
	public ArrayList<int[]> changeFileToArray(File file,
		File numFile) throws IOException {
//		System.out.println("go into change file to array");
//		System.out.println("the file "+file.getAbsolutePath() );
		if (!file.isFile()) {
			System.out.println("the file " + file.getAbsolutePath() + " is not there!");
		}

		BufferedReader fileBR = new BufferedReader(new FileReader(file));
//E		String currentLine;
		ArrayList<int[]> textWordToNum = new ArrayList<int[]>(1000);

		StringBuilder sb = new StringBuilder();
		int c = 0;
		while ((c = fileBR.read()) != -1) {
			sb.append((char) c);
		}
		sb.append(" "); //!!! This is needed to get a possible last empty line in the file.
		String[] lines = sb.toString().split("\n");

//E		while(null!=(currentLine=fileBR.readLine()))
		for (String currentLine : lines) {

//			System.out.println(currentLine);
			String[] wordsList = currentLine.split("\\s+");
//			System.out.println("line splited");
			ArrayList<String> sentWordsList = new ArrayList<String>();
			for (int i = 0; i < wordsList.length; i++) {
				if (null == stopWordsList.get((wordsList[i]))) {
					sentWordsList.add(wordsList[i]);
				}
			}
//			System.out.println("remove stop words");

			int sentSize = sentWordsList.size();
			int[] wordNumList;
			wordNumList = new int[sentSize];
//			System.out.println("wordNum list created with size "+ sentSize);

			for (int i = 0; i < sentWordsList.size(); i++) {
//				System.out.println("i is "+i);
				int wordNum;
				wordNum = this.singleWordToNum(sentWordsList.get(i));

				wordNumList[i] = wordNum;
			}
			textWordToNum.add(wordNumList);

		}
		fileBR.close();
		textWordToNum.trimToSize();
//		System.out.println("all lines read!");
		if (null != numFile) {
			BufferedWriter BR = new BufferedWriter(new FileWriter(numFile));
			for (int i = 0; i < textWordToNum.size(); i++) {
				for (int j = 0; j < textWordToNum.get(i).length; j++) {
					BR.write(textWordToNum.get(i)[j] + " ");
				}
				BR.write("\n");
			}
			BR.flush();
			BR.close();

		}
//		System.out.println("num file wrote!");
		return textWordToNum;
	}

	/**
	 * change a single word to be integer.
	 *
	 * @param word: the word to be changed according to multiWordsHash,and
	 * singleWordsList, calculate the corresponding integer for this word if the
	 * word never appears(it can happen for test files), return -1.
	 * @return
	 */
	public int singleWordToNum(String word) {
		/*
		 * a token that appears more than once
		 */

		if (null != (multiWordsHash.get(word))) {
			int[] times = (int[]) multiWordsHash.get(word);
			return times[0];
		} 
		/*
		 * a token appear only once or never appear in training data.
		 */ 
		else {
//			System.out.println("possible single words");
			int singleOrder = this.isSingle(word);

			if (-1 != singleOrder) {
//				System.out.println("single word");
				return (singleOrder + multiWordsHash.size());

			} else {
//				System.out.println("new word");
				return -1;
			}
		}
	}

	/**
	 * the tokens that only appear once are sorted in alphabet order. find the
	 * position of this token. if not find, return -1.
	 *
	 * @param word
	 * @return
	 */
	public int isSingle(String word) {

		int first = 0;
		int upto = singleWords.length - 1;
//		System.out.println("now the single word is :" +word);
		while (first <= upto) {
			int mid = (first + upto) / 2;  // Compute mid point.
			if (word.trim().compareTo(singleWords[mid].trim()) < 0) {
//		        	System.out.println("mid:"+ mid);
				upto = mid - 1;     // repeat search in bottom half.
			} else if (word.trim().compareTo(singleWords[mid].trim()) > 0) {
//		        	System.out.println("first:"+ first);
				first = mid + 1;  // Repeat search in top half.
			} else {
//		        	System.out.println("find the string");
				return mid;
				// Found it. return position
			}
		}
		return -1;    // Failed to find key
	}
	/*
	 * public static void main(String[] argcs) throws IOException { long
	 * start=System.currentTimeMillis(); wordCount counter=new wordCount();
	 * String filePath="/home/students/huiqin/gutenberg/de"; String
	 * stopWordFile="/home/students/huiqin/opusData/stopwords.de"; String
	 * outputFile="/home/students/huiqin/gutenberg/dewordcount.nostop.txt";
	 * //HashMap<String,int[]> wordsListNumHash = counter.count(filePath,
	 * stopWordFile, outputFile);
	 *
	 * wordToNum changer=new wordToNum(); File dir=new File(filePath); File[]
	 * files=dir.listFiles(); for(int i=0;i<files.length;i++) { String
	 * numFileName=files[i].getAbsolutePath().concat(".num"); File numFile=new
	 * File(numFileName); changer.changeFileToArray(files[i],numFile,
	 * wordsListNumHash, counter.stopWordsList); }
	 *
	 * long end=System.currentTimeMillis(); System.out.println("to get feature
	 * text for this pair costed "+(end-start)/1000 +" seconds"); }
	 */
}