/**
 * CRFalign: A CRF-based sentence aligner for parallel corpora.
 * 
 * Copyright (C) 2012  Eva Mujdricza-Maydt
 * Copyright (C) 2012  Huiqin Körkel-Qu
 * 
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, see <http://www.gnu.org/licenses/>.
 * 
 * Developers:
 * - Eva Mujdricza-Maydt (mujdricz@cl.uni-heidelberg.de)
 * - Huiqin Körkel-Qu (huiqin@cl.uni-heidelberg.de)
 */
package general.com.flexEdit;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.HashMap;

/** This feature is a normalized variant of the edit distance for two word lists. 
 * The distance will be computed as {@code distance = 
 * (double)goodEditCount * 2 / (sentWords1.size() + sentWords2.size());}.
 * The {@code goodEditCount} is a count of "good word pairs" in the sentences. 
 * Good word pairs have an edit distance below a word length-dependent 
 * threshold. The distance will be normalized by the length of both sentences.
 * 
 * The edit distance of two words will be computed with an implementation by
 * Chas Emerick.
 *
 * @author Chas Emerick
 * @author Eva Mujdricza-Maydt
 */
public class ExtractFlexibleEditDistance {
	
	/**
	 * A threshold which give you the minimal percentage of the required 
	 * coverage of two words to be counted for the flexible edit distance
	 * feature.
	 * E.g. if the given percentage is 0.8 and
	 * a word is 9 characters long, then a good word pair to this word is 
	 * another word which has an edit distance to the first word not bigger 
	 * than 2.
	 */
	public static double MINCOVPERCENTAGE = 0.8;
	
	private static HashMap<Integer, Integer> maximumDistanceMap;
	
	/** Computes the edit (Levensthein) distance for the given sentence-pairs.
	 * The distance will be computed as {@code distance = 
	 * (double)goodEditCount * 2 / (sentWords1.size() + sentWords2.size());}.
	 * The {@code goodEditCount} is a count of "good word pairs" in the sentences. 
	 * Good word pairs have an edit distance below a word length-dependent 
	 * threshold. The distance will be normalized by the length of both sentences.
	 * 
	 * @param sentWords1 list of words in the first sentence to compare
	 * @param sentWords2 list of words in the second sentence to compare
	 * @param minCoverPcg the percentage of minimal required coverage of two
	 *			words to be a "good word pair". The percentage should be 
	 *			in range [0.0, 1.0]. We propose minCoverPcg to be 0.8.
	 * @return the normalized edit distance for the two sentences
	 */
	public static double computeFlexibleEditDistance(
			ArrayList<String> sentWords1, ArrayList<String> sentWords2,
			double minCoverPcg) {
		
		HashMap<Integer, Integer> maxDistMap = getMaximumDistanceMap(minCoverPcg);
		
		//all words to lower case!
		String[] a1 = new String[sentWords1.size()];
		for (int a = 0; a < sentWords1.size(); a++) {
			a1[a] = sentWords1.get(a).toLowerCase();
		}
		sentWords1 = new ArrayList<String>();
		sentWords1.addAll(Arrays.asList(a1));
		
		String[] a2 = new String[sentWords2.size()];
		for (int a = 0; a < sentWords2.size(); a++) {
			a2[a] = sentWords2.get(a).toLowerCase();
		}
		sentWords2 = new ArrayList<String>();
		sentWords2.addAll(Arrays.asList(a2));
		
		double distance = 0.0;
		int wchLen1 =getWordCharacterLen(sentWords1);
		int wchLen2 =getWordCharacterLen(sentWords2);
		
		//if one of the sentences is empty, no good distance is possible
		if (sentWords1.isEmpty() || sentWords2.isEmpty()) {
				distance = 0.0;
		} 
		else {
			int goodEditCount = 0; //how many word pairs with good edit distance
			boolean[] chosenWords = new boolean[sentWords2.size()]; //initialized with false
			
			//we only want to have the shared words
			for (int i = 0; i < sentWords1.size(); i++) {
				int currWordDist = wchLen1+wchLen2+1; //initialize with an impossible big value
				int currJIndex = -1;
				String w1 = sentWords1.get(i);
				if( hasNotChosen(chosenWords)) {
				
					for (int j = 0; j < sentWords2.size(); j++) {
						if (chosenWords[j] == true) {
							continue;
						}
						String w2 = sentWords2.get(j);
						int candidateDist = getLevenshteinDistance(w1, w2);
						if (candidateDist < currWordDist) {
							currWordDist = candidateDist;
							currJIndex = j;
						}
						if (currWordDist == 0) {
							break;
						}
					}

					if (currJIndex == (-1)) {
						throw new NullPointerException("No minimal edit dist. word chosen for word " + w1);
					}
					
					//backward safety: is there any better words in the first sentence for the 
					//found word of the second sentence? If yes, drop the alignment.
					if ((currWordDist > 0)
							&& isThereABetterChoice(sentWords1, i, sentWords2.get(currJIndex), currWordDist)) {
						currJIndex = (-1);
					} else {
						//sign the index with the lowest distance value
						chosenWords[currJIndex] = true;
					}
				}
				
				//here it could be, that no good j will be found
				if (currJIndex == (-1)) {
					continue;
				}
				
				//increment goodEditCount if needed
				if (currWordDist == 0) { //if zero distance -> good
					goodEditCount++;
				}
				//if distance is bigger than zero, than it is needed, that both 
				//of the words are longer than the given threshold for the 
				//current word length specified in the maxDistMap
				else {
					int w1Len = w1.length();
					int w2Len = sentWords2.get(currJIndex).length();
				
					//both words should be longer than the edit distance between
					//them; and the distance of the words should not exceed 
					//a length-dependent threshold for both words.
					if ((w1Len > currWordDist) && (w2Len > currWordDist)
						&& (currWordDist <= maxDistMap.get(w1Len)) 
						&& (currWordDist <= maxDistMap.get(w2Len))) {
						goodEditCount++;
					}
				}
			}
			distance = (double)goodEditCount * 2 / (sentWords1.size() + sentWords2.size());
		}
		return distance;
	}
	
	/** builds a map of word length (in characters) to the allowed maximal
	 edit distance. If a word pair has a lower edit distance for their length 
	 as specified by this map, then it will be counted as a "good" word pair. 
	 The map will be built once, for lenght up to 50 characters. */
	private static HashMap<Integer, Integer> getMaximumDistanceMap(
		double minCoveragePercentage) {
		
		if (maximumDistanceMap == null) {
			
			if (minCoveragePercentage < 0.0 || minCoveragePercentage > 1.0) {
				String msg = "Give a number between 0.0 and 1.0 for a "
					+ "minimum coverage percentage of two words (given: " 
					+ minCoveragePercentage + ")";
				throw new IllegalArgumentException(msg);
			}
			
			maximumDistanceMap = new HashMap<Integer, Integer>();
			maximumDistanceMap.put(0, 0);
			//we suppose that a word will be not longer than 50 characters
			for (int len = 1; len <= 50; len++) {
				int maxDist = (int)(len - Math.round(minCoveragePercentage * len));
				maximumDistanceMap.put(len, maxDist);
			}
		}
		
		return maximumDistanceMap;
	}


	/** whether there is still any word without a specified pair. */
	private static boolean hasNotChosen(boolean[] list) {
		boolean has = false;
		
		for (int i = 0; i < list.length; i++) {
			if (list[i] == false) {
				has = true;
				break;
			}
		}
		return has;
	}
	
	/** computes only the sum of the characters without spaces. */
	private static int getWordCharacterLen(ArrayList<String> wordList) {
		int l = 0;
		for (String s : wordList) {
			l += s.length();
		}
		return l;
	}
	
	/** take always the best possible word-pair, that is, the word-pair with the
	* highest overlapping, that is, with the lowest distance. */
	private static boolean isThereABetterChoice(
		ArrayList<String> sentWords1, int index1From, 
		String foundWordFromSent2, int foundDistance) {
		
		boolean is = false;
		for (int i = index1From; i < sentWords1.size(); i++) {
			if(getLevenshteinDistance(sentWords1.get(i), foundWordFromSent2) < foundDistance) {
				is = true;
				break;
			}
		}
		return is;
	}
	
	
	/**	Edit (Levenshtein) distance in implementation by Chas Emerick.
	 * <pre>
	From an email from Chas Emerick to Michael Gilleland, 22 October 2003:
	
	Mr. Gilleland,
	
	As you may know, the Apache Jakarta Commons project had appropriated 
	your sample implementation of the Levenshtein Distance algorithm for 
	its commons-lang Java library.  While attempting to use it with two 
	very large strings, I encountered an OutOfMemoryError, due to the fact 
	that a matrix is created with the dimensions of the two strings' 
	lengths.  I know you created the implementation to go with your 
	(excellent) illustration of the algorithm, so this matrix approach 
	translates that illustration and tutorial perfectly.
	
	However, as I said, the matrix approach doesn't lend itself to getting 
	the edit distance of two large strings.  For this purpose, I modified 
	your implementation to use two single-dimensional arrays; this is 
	clearly more memory-friendly (although it probably results in some very 
	slight performance degradation when comparing smaller strings).
	
	I've submitted the modification to the maintainers of the commons-lang 
	project, and I've appended the relevant method below.
	
	Thanks!
	
	Chas Emerick
	</pre>
	 * 
	 * @param s string to compare
	 * @param t string to compare
	 * @return distance number of the strings compared
	 * @author Chas Emerick
	 * @see http://www.merriampark.com/ldjava.htm (not provided any more)
	 * @see http://mrfoo.de/archiv/1176-Levenshtein-Distance-in-Java.html 
	 * (last check: 03.08.2012)
	 * @see http://lium3.univ-lemans.fr/diarization/javadoc/fr/lium/spkDiarization/libModel/Distance.html
	 * (last check: 03.08.2012)
	 */
	private static int getLevenshteinDistance(String s, String t) {
		if (s == null || t == null) {
			throw new IllegalArgumentException("Strings must not be null");
		}

		/*
		The difference between this impl. and the previous is that, rather 
		than creating and retaining a matrix of size s.length()+1 by t.length()+1, 
		we maintain two single-dimensional arrays of length s.length()+1.  The first, d,
		is the 'current working' distance array that maintains the newest distance cost
		counts as we iterate through the characters of String s.  Each time we increment
		the index of String t we are comparing, d is copied to p, the second int[].  Doing so
		allows us to retain the previous cost counts as required by the algorithm (taking 
		the minimum of the cost count to the left, up one, and diagonally up and to the left
		of the current cost count being calculated).  (Note that the arrays aren't really 
		copied anymore, just switched...this is clearly much better than cloning an array 
		or doing a System.arraycopy() each time  through the outer loop.)
		
		Effectively, the difference between the two implementations is this one does not 
		cause an out of memory condition when calculating the LD over two very large strings.  		
		 */

		int n = s.length(); // length of s
		int m = t.length(); // length of t

		if (n == 0) {
			return m;
		} else if (m == 0) {
			return n;
		}

		int p[] = new int[n + 1]; //'previous' cost array, horizontally
		int d[] = new int[n + 1]; // cost array, horizontally
		int _d[]; //placeholder to assist in swapping p and d

		// indexes into strings s and t
		int i; // iterates through s
		int j; // iterates through t

		char t_j; // jth character of t

		int cost; // cost

		for (i = 0; i <= n; i++) {
			p[i] = i;
		}

		for (j = 1; j <= m; j++) {
			t_j = t.charAt(j - 1);
			d[0] = j;

			for (i = 1; i <= n; i++) {
				cost = s.charAt(i - 1) == t_j ? 0 : 1;
				// minimum of cell to the left+1, to the top+1, diagonally left and up +cost				
				d[i] = Math.min(Math.min(d[i - 1] + 1, p[i] + 1), p[i - 1] + cost);
			}

			// copy current distance counts to 'previous row' distance counts
			_d = p;
			p = d;
			d = _d;
		}

		// our last action in the above loop was to switch d and p, so p now 
		// actually has the most recent cost counts
		return p[n];
	}
}
