/**
 * CRFalign: A CRF-based sentence aligner for parallel corpora.
 * 
 * Copyright (C) 2012  Eva Mujdricza-Maydt
 * Copyright (C) 2012  Huiqin Körkel-Qu
 * 
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, see <http://www.gnu.org/licenses/>.
 * 
 * Developers:
 * - Eva Mujdricza-Maydt (mujdricz@cl.uni-heidelberg.de)
 * - Huiqin Körkel-Qu (huiqin@cl.uni-heidelberg.de)
 */
package objects;

import java.util.ArrayList;

/** Represents one sentence-level alignment in a source-target document pair.
 * That is a sentence or sentence chain from the source document aligned to 
 * a sentence or sentence chain from the target document.
 * <p/>
 * NOTE that "sentence-level alignments" mean that sentences will be aligned.
 * One sentence-level alignment entails one or more sentence number(s) 
 * of the source document and one or more sentence number(s) of the target 
 * document. The sentences with the given numbers are aligned in the given 
 * alignment. If a source sentence has no counterpart in the target document,
 * then on the target side, an empty list will be stored. If a target sentence has
 * no counterpart in the source document, then the source sentence number list
 * keeps empty. These two cases are so called "zero alignments". NOTE that 
 * in a zero alignment only one sentence will be bounded to "zero", even if there
 * are more sentences one after another which are aligned with no other sentences.
 *
 * @author Eva Mujdricza-Maydt
 * @lastChange 20120720
 */
public final class Alignment_SentLevel {

	//NOTE that an empty list indicates that an alignment is "zero" (i.e. no sentence)
	//e.g. sentNr_doc1 = [129], sentNr_doc2 = null
	//     the 129. sentence in doc1 has no corresponding sentence in doc2
	//NOTE that if the list is null, it means that no alignment was set
	//NOTE that at most one of the two lists is allowed to be empty.
	
	/** number of sentence(s) from the source file */
	private ArrayList<Integer> sentNr_doc1;
	/** number of sentence(s) from the target file */
	private ArrayList<Integer> sentNr_doc2;

	/** We provide a static factory for the sentence-level alignments. */
	private Alignment_SentLevel() {
		//ok
	}
	
	/** Gets a new instance for a sentence-level alignment.
	 * 
	 * @return a new sentence-level alignment instance
	 */
	public static Alignment_SentLevel newInstance() {
		return new Alignment_SentLevel();
	}

///////////////////////////////////////////////////////////////////////////////

	/** Sets the source sentence number(s) for the given alignment.
	 * <p/>
	 * NOTE that if the parameter is an empty list, then the alignment is a 
	 * zero-alignment with no counterpart for the target sentence of the 
	 * alignment.
	 * 
	 * @param sn list of source sentence numbers
	 */
	public void setSourceSentAlign(ArrayList<Integer> sn) {
		if (sn == null) {
			throw new NullPointerException("Given sentence number list is null -> not to set.");
		}
		if (this.sentNr_doc1 == null) {
			this.sentNr_doc1  = new ArrayList<Integer>();
		}
		for (Integer i : sn) {
			addToSourceSentAlign(i);
		}
	}

	/** Sets the source sentence number for the given alignment. In this 
	 * case, only one source sentence will be aligned to the target side.
	 * <p/>
	 * NOTE that if the parameter n is under a minimal value 
	 * {@link main.Constants_Main.MINIMUMSENTNUMBER}, which is set to one, then 
	 * an empty sentence number list will be generated, that is a zero alignment.
	 * 
	 * @param n source sentence number
	 */
	public void setSourceSentAlign(Integer n) {
		if (n == null) {
			throw new NullPointerException("Given sentence number is null -> not to set.");
		}
		ArrayList<Integer> sn = new ArrayList<Integer>();
		if (n >= main.Constants_Main.MINIMUMSENTNUMBER) {
			sn.add(n);
			this.sentNr_doc1 = sn;
		} else {
			//the alignment keeps empty, this indicates a "zero to something"
			//alignment
			//NOTE that a set alignment should not be null, because a null-alignment
			//means "no alignment there". On the contrary, an empty alignment means
			//zero alignment.
			setEmptySourceSentAlign();
		}
	}
	
	/** Sets the source sentence numbers for the given alignment.
	 * <p/>
	 * NOTE that it will be assumed, that the parameter sn is a String of one or 
	 * more valid Integer-values. If the String entails more Integers, then 
	 * they should be separated by a white space.
	 * <p/>
	 * NOTE that if the parameter sn is under a minimal value 
	 * {@link main.Constants_Main.MINIMUMSENTNUMBER}, which is set to one, then 
	 * a {@code NullPointerException} will be thrown.
	 * <p/>
	 * NOTE that if the parameter sn is an empty String, then 
	 * an empty sentence number list will be generated (zero alignment).
	 * 
	 * @param sn source sentence number(s) as a String
	 */
	public void setSourceSentAlign(String sn) {
		if (sn == null) {
			throw new NullPointerException("Source sentence align is null -> not to set.");
		}
		String[] al = sn.trim().split(" ");
		ArrayList<Integer> ilist = new ArrayList<Integer>(al.length);
		for (String a : al) {
			if (a.isEmpty()) {
				continue;
			}
			int n = Integer.valueOf(a);
			if (n >= main.Constants_Main.MINIMUMSENTNUMBER) {
				ilist.add(n);
			} else {
				String msg = "\nCannot set a number under " + main.Constants_Main.MINIMUMSENTNUMBER + " as a source sentence number.\n"
					+ "Wrong input: " + sn + "\n";
				throw new NullPointerException(msg);
			}
		}
		this.sentNr_doc1 = ilist;
	}

	/** Set the source side of the alignment to an empty list. This means, 
	 * that the alignment is a zero-alignment on the source side. NOTE that
	 * at most one side is allowed to be empty.
	 * 
	 * @throws IllegalStateException if the target side of the current alignment
	 *		is already set to zero
	 */
	public void setEmptySourceSentAlign() {
		if (this.hasTargetSentAlign() && this.hasEmptyTargetSentAlign()) {
			String msg = "\nThe target side of the current alignment is already set to zero.\n";
			throw new IllegalStateException(msg);
		}
		this.sentNr_doc1 = new ArrayList<Integer>(0);
	}

	/** Adds a further source sentence number to the alignment. If no source 
	 * sentence number list set, then one will be initialized with this number.
	 * 
	 * @param n sentence number 
	 */
	public void addToSourceSentAlign(Integer n) {
		if (n == null) {
			throw new NullPointerException("Given sentence number is null -> not to set.");
		}
		if (this.sentNr_doc1 == null) {
			this.sentNr_doc1 = new ArrayList<Integer>();
		}
		if (!this.sentNr_doc1.contains(n)) {
			
			if (n >= main.Constants_Main.MINIMUMSENTNUMBER) {
				this.sentNr_doc1.add(n);
			} else {
				String msg = "\nCannot set a number under " + main.Constants_Main.MINIMUMSENTNUMBER + " as a source sentence number.\n"
					+ "Wrong input: " + n + "\n";
				throw new NullPointerException(msg);
			}
			
			java.util.Collections.sort(this.sentNr_doc1);
		}
	}

	/** Checks whether the source sentence number(s) are set.
	 * 
	 * @return {@code true} if the alignment has a list of source sentence numbers,
	 *		{@code false} otherwise
	 */
	public boolean hasSourceSentAlign() {
		boolean has = true;
		if (this.sentNr_doc1 == null) {
			has = false;
		}
		return has;
	}
	
	/** Checks whether the source side of the sentence alignment is a zero
	 * alignment side. This is the case, if the list of source sentences is
	 * empty. We also allow 0 (zero) as only entry in the sentence list to indicate
	 * a zero alignment.
	 * 
	 * @return {@code true} if the source side of the alignment is a zero alignment,
	 *			{@code false} otherwise
	 */
	public boolean hasEmptySourceSentAlign() {
		boolean has = false;
		if (this.sentNr_doc1.isEmpty() 
			|| ((this.sentNr_doc1.size()== 1) && (this.sentNr_doc1.get(0) < main.Constants_Main.MINIMUMSENTNUMBER))) {
			has = true;
		}
		return has;
	}

	/** Gets the source side of the sentence alignment.
	 * 
	 * @return the list of source sentence numbers. NOTE that the list could
	 *			be empty (in case of a zero alignment).
	 * @throws NullPointerException if there is no list of source sentence 
	 *			numbers set for the current alignment 
	 */
	public ArrayList<Integer> getSourceSentAlign() {
		if (!hasSourceSentAlign()) {
			String msg = "\nNo source sentence number(s) for alignment given.\n";
			throw new NullPointerException(msg);
		}
		return this.sentNr_doc1;
	}
	
	/** Gets the source side of the sentence alignment as a String.
	 * 
	 * @return the list of source sentence numbers as a String. NOTE that the list 
	 *			could be empty (in case of a zero alignment).
	 */
	public String getSourceSentAlign_asString() {
		return getStringFromAL_int(getSourceSentAlign());
	}
	
	/** Gets the String representation of a list of Integers.
	 * 
	 * @param list list of Integers
	 * @return the String representation of the list: In our implementation,
	 *			that is the entries of the list separated by a white space character.
	 */
	private static String getStringFromAL_int(ArrayList<Integer> list) {
		StringBuilder sb = new StringBuilder();
		for (Integer i : list) {
			sb.append(i).append(" ");
		}
		return sb.toString().trim();
	}
	
	/** Sets the target sentence number(s) for the given alignment.
	 * <p/>
	 * NOTE that if the parameter is an empty list, then the alignment is a 
	 * zero-alignment with no counterpart for the source sentence of the 
	 * alignment.
	 * 
	 * @param sn list of target sentence numbers
	 */
	public void setTargetSentAlign(ArrayList<Integer> sn) {
		if (sn == null) {
			throw new NullPointerException("Target sentence align is null -> not to set.");
		}
		if (this.sentNr_doc2 == null) {
			this.sentNr_doc2  = new ArrayList<Integer>();
		}
		for (Integer i : sn) {
			addToTargetSentAlign(i);
		}
	}
	
	/** Sets the target sentence number for the given alignment. In this 
	 * case, only one target sentence will be aligned to the source side.
	 * <p/>
	 * NOTE that if the parameter n is under a minimal value 
	 * {@link main.Constants_Main.MINIMUMSENTNUMBER}, which is set to one, then 
	 * an empty sentence number list will be generated, that is a zero alignment.
	 * 
	 * @param n target sentence number
	 */
	public void setTargetSentAlign(Integer n) {
		if (n == null) {
			throw new NullPointerException("Target sentence align is null -> not to set.");
		}
		ArrayList<Integer> tn = new ArrayList<Integer>();
		if (n >= main.Constants_Main.MINIMUMSENTNUMBER) {
			tn.add(n);
			this.sentNr_doc2 = tn;
		} else {
			//the alignment keeps empty, this indicates a "zero to something"
			//alignment
			//NOTE that a set alignment should not be null, because a null-alignment
			//means "no alignment there". On the contrary, an empty alignment means
			//zero alignment.
			setEmptyTargetSentAlign();
		}
	}
	
	/** Sets the target sentence number(s) for the given alignment.
	 * <p/>
	 * NOTE that if the parameter is an empty list, then the alignment is a 
	 * zero-alignment with no counterpart for the source sentence of the 
	 * alignment.
	 * 
	 * @param sn list of target sentence numbers
	 */
	public void setTargetSentAlign(String sn) {
		if (sn == null) {
			throw new NullPointerException("Target sentence align is null -> not to set.");
		}
		String[] al = sn.trim().split(" ");
		ArrayList<Integer> ilist = new ArrayList<Integer>(al.length);
		for (String a : al) {
			if (a.isEmpty()) {
				continue;
			}
			int n = Integer.valueOf(a);
			if (n >= main.Constants_Main.MINIMUMSENTNUMBER) {
				ilist.add(n);
			} else {
				String msg = "\nCannot set a number below " + main.Constants_Main.MINIMUMSENTNUMBER + " as a target sentence number.\n"
					+ "Wrong input: " + sn + "\n";
				throw new NullPointerException(msg);
			}
		}
		this.sentNr_doc2 = ilist;
	}
	
	/** Set the target side of the alignment to an empty list. This means, 
	 * that the alignment is a zero-alignment on the target side. NOTE that
	 * at most one side is allowed to be empty.
	 * 
	 * @throws IllegalStateException if the source side of the current alignment
	 *		is already set to zero
	 */
	public void setEmptyTargetSentAlign() {
		if (this.hasSourceSentAlign() && this.hasEmptySourceSentAlign()) {
			String msg = "\nThe source side of the current alignment is already set to zero.\n";
			throw new IllegalStateException(msg);
		}
		this.sentNr_doc2 = new ArrayList<Integer>(0);
	}

	/** Adds a further target sentence number to the alignment. If no target 
	 * sentence number list set, then one will be initialized with this number.
	 * 
	 * @param n sentence number 
	 */
	public void addToTargetSentAlign(Integer n) {
		if (n == null) {
			throw new NullPointerException("Target sentence align is null -> not to set.");
		}
		if (this.sentNr_doc2 == null) {
			this.sentNr_doc2 = new ArrayList<Integer>();
		}
		if (!this.sentNr_doc2.contains(n)) {
			if (n >= main.Constants_Main.MINIMUMSENTNUMBER) {
				this.sentNr_doc2.add(n);
			} else {
				String msg = "\nCannot set a number below " + main.Constants_Main.MINIMUMSENTNUMBER + " as a target sentence number.\n"
					+ "Wrong input: " + n + "\n";
				throw new NullPointerException(msg);
			}
			java.util.Collections.sort(this.sentNr_doc2);
		}
	}

	/** Checks whether the target sentence number(s) are set.
	 * 
	 * @return {@code true} if the alignment has a list of target sentence numbers,
	 *		{@code false} otherwise
	 */
	public boolean hasTargetSentAlign() {
		boolean has = true;
		if (this.sentNr_doc2 == null) {
			has = false;
		}
		return has;
	}

	/** Checks whether the target side of the sentence alignment is a zero
	 * alignment side. This is the case, if the list of target sentences is
	 * empty. We also allow 0 (zero) as only entry in the sentence list to indicate
	 * a zero alignment.
	 * 
	 * @return {@code true} if the source side of the alignment is a zero alignment,
	 *			{@code false} otherwise
	 */
	public boolean hasEmptyTargetSentAlign() {
		boolean has = false;
		if (this.sentNr_doc2.isEmpty() || ((this.sentNr_doc2.size()== 1) && (this.sentNr_doc2.get(0) < main.Constants_Main.MINIMUMSENTNUMBER))) {
			has = true;
		}
		return has;
	}
	
	/** Gets the target side of the sentence alignment.
	 * 
	 * @return the list of target sentence numbers. NOTE that the list could
	 *			be empty (in case of a zero alignment)
	 * @throws NullPointerException if there is no list of target sentence 
	 *			numbers set for the current alignment 
	 */
	public ArrayList<Integer> getTargetSentAlign() {
		if (!hasTargetSentAlign()) {
			String msg = "\nNo target sentence number(s) for alignment given.\n";
			throw new NullPointerException(msg);
		}
		return this.sentNr_doc2;
	}

	/** Gets the target side of the sentence alignment as a String.
	 * 
	 * @return the list of target sentence numbers as a String. NOTE that the list 
	 *			could be empty (in case of a zero alignment).
	 */
	public String getTargetSentAlign_asString() {
		return getStringFromAL_int(getTargetSentAlign());
	}
	
}
