/**
 * CRFalign: A CRF-based sentence aligner for parallel corpora.
 * 
 * Copyright (C) 2012  Eva Mujdricza-Maydt
 * Copyright (C) 2012  Huiqin Körkel-Qu
 * 
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, see <http://www.gnu.org/licenses/>.
 * 
 * Developers:
 * - Eva Mujdricza-Maydt (mujdricz@cl.uni-heidelberg.de)
 * - Huiqin Körkel-Qu (huiqin@cl.uni-heidelberg.de)
 */
package objects;

/** Representation of tokens in sentences.
 *
 * @author Eva Mujdricza-Maydt
 */
public final class Word {

	/** the word itself */
	private String token;
	/** lemma of the word */
	private String lemma;
	/** part-of-speech tag of the word */
	private String posTag;
	/** number of the sentence with the current word, the first sentence has the number 1 */
	private int sentenceNumber;	
	/** number of the current word in the current sentence, the first word has the number 1 */
	private int wordInSentNumber;

	/** Creates a new instance of <code>Word</code>.
	 */
	public Word() {
		//initialize instance variables with impossible values
		token = null;	
		lemma = null;
		posTag = null;
		sentenceNumber = 0;
		wordInSentNumber = 0;
	}


///////////////////////////////////////////////////////////////////////////////


	/**
	 * sets the current <code>Word</code>
	 *
	 * @param w	current <code>Word</code>
	 */
	public void setWord(String w) {
		this.token = w;
	}

	/**
	 * Checks whether for the current word a word-String is set.
	 *
	 * @return	{@code true} if for the current word a word-String is set;
	 *		{@code false} otherwise
	 */
	private boolean hasWord() {
		boolean hasWord = true;
		if (this.token == null) {
			hasWord = false;
		}
		return hasWord;
	}

	/**
	 * gets the current <code>Word</code>
	 *
	 * @return	current <code>Word</code>
	 * @throws NullPointerException if no word is set for the given object
	 */
	public String getWord() {
		//check
		if (hasWord() == false) {
			String msg = "Token for this word is not set";
			throw new NullPointerException(msg);
		}
		return this.token;
	}
	
	/** Sets the lemma information for the current word.
	 * 
	 * @param l lemma of the current word
	 */
	public void setLemma(String l) {
		this.lemma = l;
	}

	/**
	 * Checks whether for the current word a lemma-String is set.
	 *
	 * @return	{@code true} if the lemma information is set;
	 *		{@code false} otherwise
	 */
	private boolean hasLemma() {
		boolean has = true;
		if (this.lemma == null) {
			has = false;
		}
		return has;
	}

	
	/** Gets the lemma information of the current word. 
	 *  
	 * @return lemma of the current word
	 * @throws NullPointerException if no lemma is set for the current word
	 */
	public String getLemma() {
		//check
		if (hasLemma() == false) {
			String msg = "Lemma for this word (" + this.getWord() + ") is not set";
			throw new NullPointerException(msg);
		}
		return this.lemma;
	}


	/** Sets the part-of-speech tag of the current word. 
	 * 
	 * @param p part-of-speech tag for the current word
	 */
	public void setPosTag(String p) {
		this.posTag = p;
	}

	/**
	 * Checks whether for the current word the part-of-speech tag is set.
	 *
	 * @return	{@code true} if the part-of-speech tag is set;
	 *		{@code false} otherwise
	 */
	private boolean hasPosTag() {
		if (this.posTag == null) {
			return false;
		} else {
			return true;
		}
	}

	/** Gets the part-of-speech tagof the current word. 
	 *  
	 * @return part-of-speech tag of the current word
	 * @throws NullPointerException if no part-of-speech tag is set for the current word
	 */
	public String getPosTag() {
		//check
		if (hasPosTag() == false) {
			String msg = "Part-of-speech tag for this word (" + this.getWord() + ") is not set";
			throw new NullPointerException(msg);
		}
		return this.posTag;
	}
	
	/**
	 * Sets the sentence number of the current <code>Word</code>.
	 *
	 * @param number	sentence number of the current <code>Word</code>
	 */
	public void setSentenceNr(int number) {
		this.sentenceNumber = number;
	}
	
	/**
	 * Checks whether for the current word the sentence number is set.
	 *
	 * @return	{@code true} if the sentence number is set;
	 *		{@code false} otherwise
	 */
	private boolean hasSentenceNr() {
		if (this.sentenceNumber > 0) {
			return true;
		} else {
			return false;
		}
	}

	/**
	 * Gets the sentence number of the current <code>Word</code>.
	 *
	 * @return	sentence number of the current <code>Word</code>
	 * @throws NullPointerException if no sentence number is set for the current word
	 */
	public int getSentenceNr() {
		//check
		if (hasSentenceNr() == false) {
			String msg = "\nNo sentence number for this word (" + this.getWord() + ") is set\n";
			throw new NullPointerException(msg);
		}
		return this.sentenceNumber;
	}


	/**
	 * Sets the number of the current <code>Word</code> in the sentence.
	 *
	 * @param number	number of the current <code>Word</code>
	 *			in the current sentence
	 */
	public void setWordInSentNr(int number) {
		this.wordInSentNumber = number;
	}

	/**
	 * Checks whether for the current word the word number (within the 
	 * current sentence) is set.
	 *
	 * @return	{@code true} if the word number is set;
	 *		{@code false} otherwise
	 */
	private boolean hasWordInSentNr() {
		if (wordInSentNumber > 0) {
			return true;
		} else {
			return false;
		}
	}
	
	/**
	 * Gets the number of the current <code>Word</code> in the sentence.
	 *
	 * @return	number of the current <code>Word</code> in the sentence
	 * @throws NullPointerException if no word number is set for the current word
	 */
	public int getWordInSentNr() {
		//check
		if (hasWordInSentNr() == false) {
			String msg = "\nNo sentence number for this word (" + this.getWord() + ") is set\n"
				+ "(Word number in sentence is 0 which is a not allowed value.)\n";
			throw new NullPointerException(msg);
		}
		return this.wordInSentNumber;
	}


}
