/**
 * CRFalign: A CRF-based sentence aligner for parallel corpora.
 * 
 * Copyright (C) 2012  Eva Mujdricza-Maydt
 * Copyright (C) 2012  Huiqin Körkel-Qu
 * 
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, see <http://www.gnu.org/licenses/>.
 * 
 * Developers:
 * - Eva Mujdricza-Maydt (mujdricz@cl.uni-heidelberg.de)
 * - Huiqin Körkel-Qu (huiqin@cl.uni-heidelberg.de)
 */
package utils;

import java.io.*;
import java.nio.channels.FileChannel;
import java.util.*;

/** Auxiliary methods.
 *
 * @author Eva Mujdricza-Maydt
 * @lastChange 20120720
 */
public class Helper {
	
	/** Executes a given bash script. 
	 *
	 * @param pathToApplication	path to the external application (e.g. the used
	 * part-of-speech tagger), if an application have to be started from the directory
	 * of the application; in other cases (i.e. if the script does not start
	 * an external application, or if the application can be started from the 
	 * source directory) the argument should be {@code null}.
	 * @param pathToScript	path to the given script which will be executed
	 * @param scriptFileName	the name of the given script file which will be executed
	 * @param commandLineArguments	list of command line parameters for the
	 *								script if there is any;
	 *								otherwise this argument should be {@code null}
	 * @param copyScript	whether it is needed, to copy the script into the
	 *						application folder
	 * @param printOutput	whether the user want to print the output of the
	 *						script/application to the command line
	 * @param outputFileNameWithPath output file name with path, if any needed;
	 *			if {@code null}, no output file will be written
	 * @throws IOException
	 * @throws InterruptedException
	 */
	public static void startBashScript(
		String pathToApplication,
		String pathToScript,
		String scriptFileName,
		ArrayList<String> commandLineArguments, boolean copyScript,
		boolean printOutput, String outputFileNameWithPath
		) throws IOException, InterruptedException {

		pathToApplication = normalizePath(pathToApplication);
		pathToScript = normalizePath(pathToScript);
		System.out.println("EXECUTE BASH SCRIPT " + scriptFileName);

		//script does not start an external application
		if (pathToApplication == null) {
			pathToApplication = pathToScript;
		}
		System.out.println("Path to the application: " + pathToApplication);

		String scriptFileNameWithPath = pathToApplication + scriptFileName;

		//copy the script into the application folder if needed
		boolean copied = false;
		String originalScript = pathToScript + scriptFileName;
		if (copyScript && (!hasFile(pathToApplication, scriptFileName))) {
			
			copyFile(originalScript, scriptFileNameWithPath);
			System.out.println("--> script copied");
			copied = true;
		} else {
			scriptFileNameWithPath = originalScript;
		}

		checkFileForReading(scriptFileNameWithPath);
		File scriptFile = new File(scriptFileNameWithPath);
		
		if (!scriptFile.canExecute()) {
			System.out.println("--> script made executable");
			//first arg true: set executable; second arg false: for all user
			scriptFile.setExecutable(true, false);
		}

		//compose the command list for the process
		ArrayList<String> commandList = new ArrayList<String>();
		commandList.add(scriptFile.getAbsolutePath());
		if (commandLineArguments != null) {
			for (String a : commandLineArguments) {
				if ((a == null) || a.isEmpty()) {
					continue;
				}
				commandList.add(a);
			}
		}

		//process
		ProcessBuilder processBuilder = new ProcessBuilder(commandList);
		processBuilder.directory(scriptFile.getParentFile());
		Process process = processBuilder.start();

		//prints the output of the script/application
		//writes the output in a file
		if (printOutput || outputFileNameWithPath != null) {
			BufferedWriter bw = null;
			if (outputFileNameWithPath != null) {
				bw = new BufferedWriter(new FileWriter(outputFileNameWithPath));
			}
		
			Scanner scanner = new Scanner(process.getInputStream());
			while (scanner.hasNextLine()) {
				String nl = scanner.nextLine();
				if (printOutput) {
					System.out.println(nl);
				}
				if (bw != null) {
					bw.write(nl);
					bw.newLine();
					bw.flush();
				}
			}
			scanner.close();
			
			if (bw != null) {
				bw.close();
			}
		}

		int wf = process.waitFor();
		if (wf != 0) {
			System.err.println("Process could not be finished successfully (" + wf + ")");
		}

		//remove the copied script if needed
		if (copied) {
			if (scriptFile.delete() == false) {
				System.err.println("Script " + scriptFileNameWithPath + " could not be deleted!");
			} else {
				System.out.println("--> script deleted");
			}
		}
	}

	
	/** Checks whether the given path entails the given file name.
	 * 
	 * @param path given path
	 * @param fileName name of the file expected to be in the path
	 * @return {@true} if the file is actually in the given path, {@code false}
	 *			otherwise
	 */
	public static boolean hasFile(String path, String fileName) {

		File dir = new File(path);
		if (!dir.exists()) {
			String msg = "\nNo directory " + path + "\n";
			throw new NullPointerException(msg);
		}
		File[] files = dir.listFiles();

		boolean hasFile = false;
		for (File f : files) {
			if (f.getName().equals(fileName)) {
				hasFile = true;
				break;
			}
		}
		return hasFile;
	}

	
	/** Copies a file.
	 * 
	 * @param finName name of the file to copy, with path
	 * @param foutName the target name of the file, with path
	 * @throws IOException 
	 */
	public static void copyFile(String finName, String foutName)
		throws IOException {

		File fin = new File(finName); 
		if (!fin.exists()) {
			String msg = "\nFile to copy does not exists: \"" + finName + "\"\n";
			throw new NullPointerException(msg);
		}
		
		File fout = new File(foutName);
		FileChannel inChannel  = new FileInputStream(fin).getChannel();
		FileChannel outChannel = null;
		try {
			outChannel = new FileOutputStream(fout, false).getChannel();
		} catch (IOException ioe) {
			String msg = "\nCannot create output channel for output file \"" + foutName + "\"\n";
			System.err.println(msg);
			ioe.printStackTrace();
			System.exit(2);
		}
		try {
			inChannel.transferTo(0, inChannel.size(), outChannel);
			inChannel.close();
			outChannel.close();
		} catch (IOException ioe)  {
			String msg = "\nCannot transfer in channel between \"" + finName + "\" and \"" + foutName + "\"\n";
			System.err.println(msg);
			ioe.printStackTrace();
			System.exit(2);
		}
		
		finally {
			if(inChannel != null) {
				try {
					inChannel.close();
				} catch(IOException ex) {
					ex.printStackTrace();
					System.exit(2);
				}
			}
			if (outChannel != null) {
				try {
					outChannel.close();
				} catch(IOException ex) {
					ex.printStackTrace();
					System.exit(2);
				}
			}
		}
	}
	

	/** Gets the sorted array of Strings in a {@code Set}. 
	 * 
	 * @param keys a {@code Set} of Strings, e.g. keys from a {@code Map}
	 * @return the sorted array of the input elements
	 */
	public static String[] sortKeys_str(Set<String> keys) {
		String[] retval = keys.toArray(new String[keys.size()]);
		Arrays.sort(retval);
		return retval;
	}

	/** Gets the sorted array of Strings in a {@code Collection}. 
	 * 
	 * @param values a {@code Collection} of Strings, e.g. values from a {@code Map}
	 * @return the sorted array of the input elements
	 */
	public static String[] sortValues_str(Collection<String> values) {
		String[] retval = values.toArray(new String[values.size()]);
		Arrays.sort(retval);
		return retval;
	}
	
	
	/** Gets the sorted list of Strings in an {@code ArrayList}.
	 * 
	 * @param str list of Strings
	 * @return sorted list of Strings
	 */
	public static ArrayList<String> sortAL_str(ArrayList<String> str) {
		String[] sArray = str.toArray(new String[str.size()]) ;
		Arrays.sort(sArray);
		ArrayList<String> sortedAL = new ArrayList<String>();
		sortedAL.addAll(Arrays.asList(sArray));
		return sortedAL;
	}
	
	
	/** Extracts the file names of the given extension, within a given path. 
	 * 
	 * @param pathToFiles path to the files with the given extension (ending)
	 * @param ending extension of the files you want collect
	 * @return the list of the files with the given extension within the given
	 *			path
	 */
	public static ArrayList<String> extractFileNamesByEnding (
		String pathToFiles, String ending) {

		File f = new File(pathToFiles);
		if (!f.exists()) {
			String msg = "There is no path \"" + pathToFiles + "\"\n";
			throw new NullPointerException(msg);
		}
		File[] dir = f.listFiles(new FileFilterByEnding(ending));

		if (dir == null) {
			try {
			String msg = "There is no appropriate file with ending \""
				+ ending + "\" in the folder \"" + (new File(pathToFiles)).getCanonicalPath() + "\"\n";
			throw new NullPointerException(msg);
			} catch (IOException e) {
				
				System.exit(2);
			}
		}

		ArrayList<String> relevantFiles = new ArrayList<String>();
		for (File file : dir) {
			String fn = file.getName();
			if (fn.endsWith(ending)) {
				relevantFiles.add(fn);
			}
		}
		return sortAL_str(relevantFiles);
	}
	
	
	/** Extracts the file names of the given extension, within a given path. 
	 * The file names will be listed without the extension.
	 *  
	 * @param pathToFiles path to the files with the given extension (ending)
	 * @param ending extension of the files you want collect
	 * @return the list of the files with the given extension within the given
	 *			path, listed without their extension
	 */
	public static ArrayList<String> extractFileNamesByEnding_woEnding (
		String pathToFiles, String ending) {

		ArrayList<String> relevantFilesWithEnding = extractFileNamesByEnding(pathToFiles, ending);
		ArrayList<String> relevantFiles = new ArrayList<String>(relevantFilesWithEnding.size());
		for (String rel : relevantFilesWithEnding) {
			relevantFiles.add(changeFileNameExtension(rel, ending, ""));
		}
		return sortAL_str(relevantFiles);
	}


	/** Checks whether the given file name is ready to be opened for reading.
	 * If this is the case, nothing happens. Otherwise 
	 * {@code FileNotFoundException} or {@code IOException} will be thrown.
	 * 
	 * @param fileName given file name with path
	 * @throws FileNotFoundException if the given file name does not exist, or 
	 *			it is not a file
	 * @throws IOException if the file with the given name cannot be read
	 */
	public static void checkFileForReading(String fileName)
		throws FileNotFoundException, IOException {

		File fromFile = new File(fileName);

		if (!fromFile.exists()) {
			String msg = "\nThere is no file with file name \"" + fileName + "\".\n";
			throw new FileNotFoundException(msg);
		}

		if (!fromFile.isFile()) {
			String msg = "\nThere is no normal file with file name \"" + fileName + "\".\n";
			throw new FileNotFoundException(msg);
		}

		if (!fromFile.canRead()) {
			String msg = "\nFile \"" + fileName + "\" cannot be read.\n";
			throw new IOException(msg);
		}
		
	}

	
	/** Checks whether the given file name is ready to be opened for writing.
	 * If this is the case, nothing happens. Otherwise an {@code IOException}
	 * will be thrown.
	 * 
	 * @param toFileName given file name with path
	 * @throws IOException if the given file name is not ready for be opened for
	 *						writing.
	 */
	public static void checkFileForWriting(String toFileName)
		throws IOException {

		File toFile = new File(toFileName);
		if (toFile.exists()) {
			if (!toFile.canWrite()) {
				String msg = "\nDestination file is unwriteable: " + toFileName + "\n";
				throw new IOException(msg);
			}
		}
	}

	
	/** Changes the extension of the file name.
	 * 
	 * @param inputFileName given file name with inputExtension
	 * @param inputExtension the extension of the given file name
	 * @param outputExtension the desired extension for the file
	 * @return the file name with the outputExtension
	 */
	public static String changeFileNameExtension(
		String inputFileName, String inputExtension, String outputExtension)
		{

		String outputFileName = null;
		if (inputFileName ==  null) {
			String msg = "\nNo input file for changing the file name extension.\n";
			throw new NullPointerException(msg);
		}
		if (!inputFileName.endsWith(inputExtension)) {
			String msg = "\nThe given filename " + inputFileName + " does not ends with " + inputExtension + "\n";
			throw new NullPointerException(msg);
		} else {
			outputFileName = inputFileName.substring(
					0, (inputFileName.length()-inputExtension.length())).concat(outputExtension);
		}
		
		return outputFileName;
	}

	
	/**
	 * Gets a normalized form of a path name.
	 * If a path is not absolute, the absolute file name will be taken.
	 * If a path name does not end with a slash, this will be supplemented.
	 * If there are backslashes in the path name, all of them will be replaced
	 * with slashes.
	 * 
	 * @param path a path name, absolute or relative, with or without a tailing
	 *				slash at the end
	 * @return	normalized absolute path name, with slashes as directory
	 *			separator, and a slash at the end of the path name
	 */
	public static String normalizePath(String path) {
		
		if (path.indexOf(main.Constants_Main.BACKSLASH) > -1) {
			path = path.replace(main.Constants_Main.BACKSLASH, main.Constants_Main.SLASH);
		}
		
		try {
			File p = new File(path).getCanonicalFile();
			p = null; //not used later
		} catch (IOException ioe) {
			String msg = "\nThe given path \"" + path + "\" could not be normalized.\n";
			System.err.println(msg);
			System.err.println(ioe);
			ioe.printStackTrace();
			System.exit(2);
		}
		
		if (!path.endsWith(main.Constants_Main.SLASH)) {
			path = path.concat(main.Constants_Main.SLASH);
		}
		return path;
	}
	
	
	/** Creates a given subfolder for a given path, if the subfolder not exists.
	 * 
	 * @param path given path name
	 * @param subfolder required subfolder name
	 * @return {@code true}, if the subfolder could be created, {@code false}, 
	 *			if the subfolder is already exists, or the subfolder could not
	 *			be created
	 */
	public static boolean createSubfolderIfNeeded(String path, String subfolder) {
		boolean success = false;
		path = normalizePath(path);
		File f = new File(path + subfolder);
		if (!f.isDirectory()) {
			success = f.mkdir();
		}
		return success;
	}
}

/**
 * Filters a given file name by the file extension.
 *
 * @author Eva Mujdricza-Maydt
 */
class FileFilterByEnding implements FileFilter {

	private String ending = null;

	/**
	 * Creates a new instance of {@code FileFilterByEnding}.
	 *
	 * @param fileEnding	ending of the name of files that should be collected
	 */
	public FileFilterByEnding(String fileEnding) {
		this.ending = fileEnding;
	}


	/**
	 * Tests whether the specified pathname should be included
	 * in a pathname list.
	 *
	 * @param  pathname	the pathname to be tested
	 * @return	{@code true} if and only if pathname
	 *		with a given ending (captured from the constructor)
	 *		should be included
	 */
	@Override
	public boolean accept(File fname) {
		String name = fname.getName();
		if (name.endsWith(this.ending)) {
			return true;
		} 
		return false;
	}
}